/*
 * Copyright 2016 Capital One Services, LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __TERADATA_EXPORT_H
#define __TERADATA_EXPORT_H

#include <connection.h>
#include <schema.h>


using namespace teradata::client::API;

class TeradataExport {
    bool connected;
    Schema* dynamic_schema;
    Connection* conn;
  public:
    int connection_status;
    char* error_msg;
    TD_ErrorType error_type;

    TeradataExport(std::string server, std::string username, std::string password);
    ~TeradataExport();

    void initiate();
    void get_schema();
    void add_attribute(int attr_name, char* attr_value);
    void add_attribute(int attr_name, int attr_value);
    int get_one(unsigned char** row, int* length);
    int get_buffer(unsigned char** row, int* length);
    std::vector<Column*> get_columns();
    void close();
};

#endif
